from types import SimpleNamespace

# We cannot use Enum here because `mypy` doesn't like it when we format those routes
# into strings. https://github.com/python/mypy/issues/15269
Routes = SimpleNamespace(
    healthcheck="healthcheck",
    username_login="login",
    api_key_login="login/api_key",
    get_token="get-token",
    current_user="current_user",
)



from typing import Optional

from pydantic import UUID4, BaseModel, Field, ValidationInfo, field_validator

from galileo_core.schemas.protect.action import Action


class Stage(BaseModel):
    name: str = Field(description="Name of the stage. Must be unique within the project.")
    project_id: UUID4 = Field(description="ID of the project to which this stage belongs.")
    description: Optional[str] = Field(
        description="Optional human-readable description of the goals of this guardrail.", default=None
    )
    action: Optional[Action] = Field(
        description="An optional action (kill switch) to take that supersedes all ruleset validations.", default=None
    )
    action_enabled: bool = Field(
        description="Whether the action is enabled. If False, the action will not be applied.", default=False
    )

    @field_validator("action_enabled", mode="before")
    def validate_action_enabled(cls, value: bool, info: ValidationInfo) -> bool:
        if value and info.data.get("action") is None:
            raise ValueError("Action cannot be enabled if action is not set.")
        return value



from http.client import HTTPException
from typing import Any, Callable, Dict
from urllib.parse import urljoin

from pydantic import BaseModel, HttpUrl, SecretStr
from requests import Response

from galileo_core.constants.http_headers import HttpHeaders


class ApiClient(BaseModel):
    host: HttpUrl
    jwt_token: SecretStr

    @property
    def auth_header(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.jwt_token.get_secret_value()}"}

    @staticmethod
    def validate_response(response: Response) -> None:
        if not response.ok:
            raise HTTPException(
                "Something didn't go quite right. The api returned a non-ok status "
                f"code {response.status_code}. Error message: + {response.text}."
            ) from None

    @staticmethod
    def make_request(
        request_method: Callable[..., Response], base_url: str, endpoint: str, *args: Any, **kwargs: Any
    ) -> Any:
        response = request_method(urljoin(base_url, endpoint), *args, **kwargs)
        ApiClient.validate_response(response)
        return response.json()

    def request(
        self,
        method: Callable[..., Response],
        path: str,
        content_headers: Dict[str, str] = HttpHeaders.json(),
        **kwargs: Any,
    ) -> Any:
        return ApiClient.make_request(
            request_method=method,
            base_url=self.host.unicode_string(),
            endpoint=path,
            headers={**content_headers, **self.auth_header},
            **kwargs,
        )



from typing import Optional, Sequence

from pydantic import BaseModel, Field, field_validator

from galileo_core.schemas.protect.action import Action, BaseAction, PassthroughAction
from galileo_core.schemas.protect.rule import Rule


class Ruleset(BaseModel):
    rules: Sequence[Rule] = Field(
        description="List of rules to evaluate. Atleast 1 rule is required.", default_factory=list, min_length=1
    )
    action: Action = Field(description="Action to take if all the rules are met.", default_factory=PassthroughAction)
    description: Optional[str] = Field(description="Description of the ruleset.", default=None)
    timeout: Optional[int] = Field(
        description="Time in nano-seconds for all the rules to be evaluated. If not provided, will evaluate all the rules to completion.",
        default=None,
    )

    @field_validator("action", mode="before")
    def validate_action(cls, value: Optional[BaseAction]) -> BaseAction:
        if not value:
            value = PassthroughAction()
        return value



from enum import Enum
from typing import Optional

from pydantic import BaseModel

from galileo_core.schemas.shared.metric import MetricValueType


class MetricComputationStatus(str, Enum):
    SUCCESS = "SUCCESS"
    TIMEOUT = "TIMEOUT"
    FAILED = "FAILED"
    ERROR = "ERROR"


class MetricComputation(BaseModel):
    value: MetricValueType = None
    execution_time: Optional[float] = None
    status: Optional[MetricComputationStatus] = None
    error_message: Optional[str] = None



from datetime import timedelta
from functools import cached_property
from typing import Dict, List, Optional, Sequence, Set

from pydantic import UUID4, BaseModel, ConfigDict, Field, computed_field

from galileo_core.schemas.protect.payload import Payload
from galileo_core.schemas.protect.rule import Rule
from galileo_core.schemas.protect.ruleset import Ruleset


class Request(BaseModel):
    project_id: UUID4
    payload: Payload = Field(description="Payload to be processed.")
    rulesets: Sequence[Ruleset] = Field(
        default_factory=list,
        description="Rulesets to be applied to the payload.",
        validation_alias="prioritized_rulesets",
    )
    timeout: float = Field(
        default=timedelta(minutes=5).total_seconds(),
        description="Optional timeout for the guardrail execution in seconds. This is not the timeout for the request. If not set, a default timeout of 5 minutes will be used.",
    )
    metadata: Optional[Dict[str, str]] = Field(
        default=None,
        description="Optional additional metadata. This will be echoed back in the response.",
    )
    headers: Optional[Dict[str, str]] = Field(
        default=None,
        description="Optional additional HTTP headers that should be included in the response.",
    )

    model_config = ConfigDict(populate_by_name=True)

    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore[misc]
    @cached_property
    def rules(self) -> List[Rule]:
        rules: List[Rule] = []
        for ruleset in self.rulesets:
            rules.extend(ruleset.rules)
        return rules

    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore[misc]
    @cached_property
    def metrics(self) -> Set[str]:
        metrics_to_compute = []
        for rule in self.rules:
            metrics_to_compute.append(rule.metric)
        return set(metrics_to_compute)

    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore[misc]
    def timeout_ns(self) -> float:
        return self.timeout * 1e9



from enum import Enum
from typing import Dict


class HttpHeaders(str, Enum):
    accept = "accept"
    content_type = "Content-Type"
    application_json = "application/json"

    @staticmethod
    def accept_json() -> Dict[str, str]:
        return {HttpHeaders.accept: HttpHeaders.application_json}

    @staticmethod
    def content_type_json() -> Dict[str, str]:
        return {HttpHeaders.content_type: HttpHeaders.application_json}

    @staticmethod
    def json() -> Dict[str, str]:
        return {**HttpHeaders.accept_json(), **HttpHeaders.content_type_json()}



from functools import cached_property
from typing import Optional

from pydantic import BaseModel, Field, computed_field, model_validator


class Payload(BaseModel):
    input: Optional[str] = Field(default=None, description="Input text to be processed.")
    output: Optional[str] = Field(default=None, description="Output text to be processed.")

    @model_validator(mode="after")
    def ensure_input_or_output(self) -> "Payload":
        if not self.input and not self.output:
            raise ValueError("Either input or output must be set.")
        return self

    # https://github.com/python/mypy/issues/1362
    @computed_field  # type: ignore[misc]
    @cached_property
    def text(self) -> str:
        if self.output:
            return self.output
        elif self.input:
            return self.input
        else:
            raise ValueError("Either input or output must be set.")



from getpass import getpass
from pathlib import Path
from time import time
from typing import Optional
from urllib.parse import urljoin
from webbrowser import open_new_tab

from jwt import decode as jwt_decode
from pydantic import (
    HttpUrl,
    SecretStr,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic_core import Url
from pydantic_settings import BaseSettings, SettingsConfigDict
from requests import get, post

from galileo_core.constants.routes import Routes
from galileo_core.helpers.api_client import ApiClient
from galileo_core.helpers.logger import logger


class GalileoConfig(BaseSettings):
    console_url: HttpUrl

    # User auth details.
    username: Optional[str] = None
    password: Optional[SecretStr] = None
    api_key: Optional[SecretStr] = None
    jwt_token: Optional[SecretStr] = None
    current_user: Optional[str] = None

    # Config file for this project.
    config_filename: str = "core-config.json"

    model_config = SettingsConfigDict(
        # Allow loading from environment variables.
        env_prefix="GALILEO_",
        # Ignore unknown fields when loading from a config file.
        extra="ignore",
        # Validate fields at assignment.
        validate_assignment=True,
    )

    @computed_field
    def api_url(self) -> Url:
        # Local dev.
        if self.console_url.host in ["localhost", "127.0.0.1"]:
            api_url = "http://localhost:8088"
        else:
            api_url = self.console_url.unicode_string().replace("console", "api")
        return Url(api_url)

    def get_token_from_ui(self) -> str:
        token_url = urljoin(str(self.console_url), Routes.get_token)
        logger.info(f"🔐 Opening {token_url} to generate a new token.")
        try:
            open_new_tab(token_url)
        except Exception:
            pass
        finally:
            print(f"Go to {token_url} to generate a new token.")
            return getpass("🔐 Enter your token:")

    def login(self) -> None:
        token_data = {}
        if self.api_key:
            logger.debug("Logging in with API key.")
            token_data = ApiClient.make_request(
                post,
                base_url=str(self.api_url),
                endpoint=Routes.api_key_login,
                data=dict(api_key=self.api_key.get_secret_value()),
            )
        elif self.username and self.password:
            logger.debug("Logging in with username and password.")
            token_data = ApiClient.make_request(
                post,
                base_url=str(self.api_url),
                endpoint=Routes.username_login,
                data=dict(username=self.username, password=self.password.get_secret_value(), auth_method="email"),
            )
        else:
            logger.debug("No credentials found. Attempting to log in with token.")

        if (jwt_token := token_data.get("access_token")) is None:
            logger.debug("No credentials found. Attempting to log in with token.")
            jwt_token = self.get_token_from_ui()
        self.jwt_token = SecretStr(jwt_token)
        logger.debug("Logged in successfully.")
        self.current_user = self.api_client.request(get, path=Routes.current_user).get("email")
        self.write()
        print(f"👋 You have logged into 🔭 Galileo ({self.console_url}) as {self.current_user}.")

    def refresh_jwt_token(self) -> None:
        """Refresh token if not present or expired."""
        # Check to see if our token is expired before making a request
        # and refresh token if it's expired
        if self.jwt_token:
            claims = jwt_decode(self.jwt_token.get_secret_value(), options={"verify_signature": False})
            if claims.get("exp", 0) < time():
                self.login()

    @property
    def config_file(self) -> Path:
        return Path.home().joinpath(".galileo", self.config_filename)

    @field_validator("console_url", mode="before")
    def http_url(cls, value: str) -> str:
        if value and not (value.startswith("https") or value.startswith("http")):
            value = f"https://{value}"
        return value

    @model_validator(mode="after")
    def validate_api_url(self) -> "GalileoConfig":
        ApiClient.make_request(get, base_url=str(self.api_url), endpoint=Routes.healthcheck)
        return self

    @field_serializer("password", "jwt_token", "api_key", when_used="json-unless-none")
    def serialize_secrets(self, value: SecretStr) -> str:
        return value.get_secret_value()

    def write(self) -> None:
        self.config_file.parent.mkdir(exist_ok=True)
        self.config_file.write_text(self.model_dump_json(exclude_none=True))

    @property
    def api_client(self) -> ApiClient:
        self.refresh_jwt_token()
        if not self.jwt_token:
            raise ValueError("No token set. Please log in.")
        return ApiClient(host=self.api_url, jwt_token=self.jwt_token)

    def reset(self) -> None:
        # Reset credentials.
        self.username = None
        self.password = None
        self.api_key = None
        self.jwt_token = None
        self.current_user = None

    def logout(self) -> None:
        self.reset()
        self.write()
        print(f"👋 You have logged out of 🔭 Galileo ({self.console_url}).")
